// Copyright 2017 Fortanix, Inc. All rights reserved.
//
// This example demonstrates how to connect to the SDKMS server as an
// application using an API key and perform encryption or decryption
// using an AES key in GCM mode.

package com.fortanix.sdkms.v1.examples;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Base64;
import java.util.List;
import org.apache.commons.cli.*;
import org.apache.commons.io.IOUtils;

// API library imports.
import com.fortanix.sdkms.v1.ApiClient;
import com.fortanix.sdkms.v1.ApiException;
import com.fortanix.sdkms.v1.Configuration;
import com.fortanix.sdkms.v1.api.AuthenticationApi;
import com.fortanix.sdkms.v1.api.EncryptionAndDecryptionApi;
import com.fortanix.sdkms.v1.api.SecurityObjectsApi;
import com.fortanix.sdkms.v1.auth.ApiKeyAuth;
import com.fortanix.sdkms.v1.model.AuthResponse;
import com.fortanix.sdkms.v1.model.CryptMode;
import com.fortanix.sdkms.v1.model.DecryptRequest;
import com.fortanix.sdkms.v1.model.DecryptResponse;
import com.fortanix.sdkms.v1.model.EncryptRequest;
import com.fortanix.sdkms.v1.model.EncryptResponse;
import com.fortanix.sdkms.v1.model.KeyObject;
import com.fortanix.sdkms.v1.model.ObjectType;

// For convenience, you might want to just use the following imports instead:
//import com.fortanix.sdkms.v1.*;
//import com.fortanix.sdkms.v1.api.*;
//import com.fortanix.sdkms.v1.auth.*;
//import com.fortanix.sdkms.v1.models.*;

public class AesGcmCrypto {

    private enum CryptoOp {
        ENCRYPT,
        DECRYPT
    }

    private boolean debug;
    private boolean debug_tls;
    private String appId;
    private String server;
    private String truststore;
    private String apiKey;
    private String keyId;
    private String keyName;
    private String ivIn;
    private String ivOut;
    private String adIn;
    private String tagIn;
    private String tagOut;
    private String bearerToken;
    private CryptoOp direction;
    private Integer tagLen;

    public void connect() {
        ApiClient client = new ApiClient();
        // Set the name of the server to talk to.
        client.setBasePath(server);

        // This optionally enables verbose logging in the API library.
        client.setDebugging(debug);

        // The default ApiClient (and its configured authorization) will be
        // used for constructing the specific API objects, such as
        // EncryptionAndDecryptionApi and SecurityObjectsApi.
        Configuration.setDefaultApiClient(client);

        // If you need a trust store for the server's certificate, you can
        // configure it here or via -Djavax.net.ssl.trustStore= on the
        // java command line.
        if (truststore != null) {
            System.setProperty("javax.net.ssl.trustStore", truststore);
        }

        // This optionally enables very verbose logging in the Java network
        // libraries.
        if (debug_tls) {
            System.setProperty("javax.net.debug", "all");
        }

        // When authenticating as an application, the API Key functions as
        // the entire HTTP basic auth token.
        client.setBasicAuthString(apiKey);

        // Acquire a bearer token to use for other APIs.
        try {
            AuthResponse response = new AuthenticationApi().authorize();
            bearerToken = response.getAccessToken();
            if (debug) {
                System.err.printf("Received Bearer token %s\n", bearerToken);
            }

            // Configure the client library to use the bearer token.
            ApiKeyAuth bearerAuth =
                (ApiKeyAuth) client.getAuthentication("bearerToken");
            bearerAuth.setApiKey(bearerToken);
            bearerAuth.setApiKeyPrefix("Bearer");
        } catch (ApiException e) {
            System.err.println("Unable to authenticate: " + e.getMessage());
            System.exit(1);
        }
    }

    public KeyObject findKey() throws ApiException {
        if (keyName != null) {
            // This filters by key name. Since names are unique, this will
            // return either 0 or 1 object.
            List<KeyObject> keys =
                new SecurityObjectsApi().getSecurityObjects(keyName, null, null);
            if (keys.size() == 0) {
                throw new ApiException("No key found with name " + keyName);
            } else {
                return keys.get(0);
            }
        } else {
            // This returns the security object with the given ID.
            return new SecurityObjectsApi().getSecurityObject(keyId);
        }
    }

    public void performCrypto() throws ApiException, IOException {
        // Find a key by name or ID. You could skip this step if you already
        // know the ID of the key you wish to encrypt or decrypt with and
        // the key type.
        KeyObject key = findKey();
        if (key.getObjType() != ObjectType.AES) {
            throw new ApiException("Requested key had type " +
                                   key.getObjType().toString() +
                                   ". Please run this test with an AES key");
        }

        // This program gets the input data from stdin.
        byte[] input = IOUtils.toByteArray(System.in);

        byte[] output;

        if (direction == CryptoOp.ENCRYPT) {
            byte[] ivBytes = null;
            // Get initialization value if provided. If it's not provided,
            // SDKMS will generate a random IV for us and return it.
            if (ivIn != null) {
                ivBytes = readFileAsBytes(ivIn);
            }

            // Authentication data is optional for AES-GCM.
            byte[] adBytes = null;
            if (adIn != null) {
                adBytes = readFileAsBytes(adIn);
            }

            // For symmetric encryption with GCM and CCM modes, the plaintext,
            // algorithm, cipher mode, optional iv, optional ad, and tagLen
            // values must be supplied in the request body.
            EncryptRequest request = new EncryptRequest()
                .plain(input)
                .alg(key.getObjType())
                .mode(CryptMode.GCM)
                .tagLen(tagLen);
            if (ivBytes != null) {
                request.setIv(ivBytes);
            }
            if (adBytes != null) {
                request.setAd(adBytes);
            }

            EncryptResponse response =
                // The ID of the key is provided in the API call.
                new EncryptionAndDecryptionApi().encrypt(key.getKid(), request);
            output = response.getCipher();
            if (ivOut != null) {
                writeFileAsBytes(ivOut, response.getIv());
            }
            writeFileAsBytes(tagOut, response.getTag());
        } else {
            // Block cipher modes GCM and CCM require an IV to be able to
            // decrypt, and a tag to be able to authenticate the encrypted data.
            byte[] ivBytes = readFileAsBytes(ivIn);
            byte[] tagBytes = readFileAsBytes(tagIn);

            // Authentication data is optional for AES-GCM, but if it was
            // supplied during encryption, the same AD must be supplied
            // at decryption.
            byte[] adBytes = null;
            if (adIn != null) {
                adBytes = readFileAsBytes(adIn);
            }

            // For symmetric decryption with GCM and CCM modes, the ciphertext,
            // algorithm, cipher mode, iv, tag and optional ad must be
            // supplied in the request body.
            DecryptRequest request = new DecryptRequest()
                .cipher(input)
                .alg(key.getObjType())
                .mode(CryptMode.GCM)
                .iv(ivBytes)
                .tag(tagBytes);
            if (adBytes != null) {
                request.setAd(adBytes);
            }
            DecryptResponse response =
                // The ID of the key is provided in the API call.
                new EncryptionAndDecryptionApi().decrypt(key.getKid(), request);
            output = response.getPlain();
        }
        // And then we output to stdout.
        IOUtils.write(output, System.out);
    }

    public void logout() {
        if (bearerToken != null) {
            // It is a good idea to terminate the session when you are done
            // using it. This minimizes the window of time in which an attacker
            // could steal your bearer token and use it.
            try {
                new AuthenticationApi().terminate();
            } catch (ApiException e) {
                System.err.println("Error logging out: " + e.getMessage());
            }
            bearerToken = null;
        }
    }

    public String cipherModes() {
        String modes = null;
        for (CryptMode mode : CryptMode.values()) {
            if (modes == null) {
                modes = mode.toString();
            } else {
                modes += ", " + mode.toString();
            }
        }
        return modes;
    }

    public byte[] readFileAsBytes(String filename) throws IOException {
        FileInputStream in = null;
        try {
            in = new FileInputStream(filename);
            return IOUtils.toByteArray(in);
        } finally {
            if (in != null) {
                in.close();
            }
        }
    }

    public void writeFileAsBytes(String filename, byte[] bytes)
        throws IOException {
        FileOutputStream out = null;

        try {
            out = new FileOutputStream(filename);
            IOUtils.write(bytes, out);
        } finally {
            if (out != null) {
                out.close();
            }
        }
    }

    protected void usage(Options options) {
        HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp("AesGcmCrypto",
                            "Encrypts or decrypts data using keys in " +
                            "an SDKMS server. Input is read from stdin " +
                            "and written to stdout", options,
                            "");
    }

    protected void help(String msg) {
        System.err.println(msg);
        System.err.println("Run AesGcmCrypto --help for usage information");
    }

    public void parseCmdline(String[] args) {
        Options options = new Options();
        options.addOption(Option.builder("s")
                          .longOpt("server")
                          .hasArg()
                          .type(String.class)
                          .desc("SDKMS server endpoint (e.g. " +
                                "https://apps.smartkey.io")
                          .build());

        options.addOption(Option.builder("k")
                          .longOpt("api-key")
                          .required(true)
                          .hasArg()
                          .type(String.class)
                          .desc("API key for this application")
                          .build());

        options.addOption(Option.builder("K")
                          .longOpt("key-id")
                          .hasArg()
                          .type(String.class)
                          .desc("Key ID of the cryptographic key to use")
                          .build());

        options.addOption(Option.builder("n")
                          .longOpt("key-name")
                          .hasArg()
                          .type(String.class)
                          .desc("Key name of the cryptographic key to use")
                          .build());

        options.addOption(Option.builder("e")
                          .longOpt("encrypt")
                          .desc("Perform encryption")
                          .build());

        options.addOption(Option.builder("E")
                          .longOpt("decrypt")
                          .desc("Perform decryption")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("iv-in")
                          .hasArg()
                          .type(String.class)
                          .desc("Name of file containing block cipher " +
                                "input initialization value")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("iv-out")
                          .hasArg()
                          .type(String.class)
                          .desc("Name of the file to write the block cipher " +
                                "initialization value to")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("tag-len")
                          .hasArg()
                          .type(Number.class)
                          .desc("Length of authentication tag (valid lengths " +
                                "are 128, 120, 112, 104, 96, 64 or 32)")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("tag-out")
                          .hasArg()
                          .type(String.class)
                          .desc("Name of the file to write the cryptographic " +
                                "tag data to")
                          .build());
        options.addOption(Option.builder()
                          .longOpt("tag-in")
                          .hasArg()
                          .type(String.class)
                          .desc("Name of the file to read the cryptographic " +
                                "tag data from")
                          .build());

        options.addOption(Option.builder()
                          .longOpt("ad-in")
                          .hasArg()
                          .type(String.class)
                          .desc("Optional name of the file to read " +
                                "input authentication data from")
                          .build());

        options.addOption(Option.builder("t")
                          .longOpt("truststore")
                          .hasArg()
                          .type(String.class)
                          .desc("Java truststore containing server certificate")
                          .build());

        options.addOption(Option.builder("d")
                          .longOpt("debug")
                          .desc("Enable client API library debugging")
                          .build());

        options.addOption(Option.builder("D")
                          .longOpt("debug-tls")
                          .desc("Enable Java TLS debugging")
                          .build());

        options.addOption(Option.builder("h")
                          .longOpt("help")
                          .desc("Get usage information and exit")
                          .build());

        CommandLineParser parser = new DefaultParser();
        try {
            CommandLine cmdline = parser.parse(options, args);
            if (cmdline.hasOption("help")) {
                usage(options);
                System.exit(0);
            }

            debug = cmdline.hasOption("debug");
            debug_tls = cmdline.hasOption("debug-tls");
            appId = cmdline.getOptionValue("app-id");
            apiKey = cmdline.getOptionValue("api-key");

            if (cmdline.hasOption("server")) {
                server = cmdline.getOptionValue("server");
            } else {
                server = "https://apps.smartkey.io";
            }

            if (cmdline.hasOption("truststore")) {
                truststore = cmdline.getOptionValue("truststore");
            }

            if ((!cmdline.hasOption("encrypt") &&
                 !cmdline.hasOption("decrypt")) ||
                (cmdline.hasOption("encrypt") &&
                 cmdline.hasOption("decrypt"))) {
                help("You must specify either --encrypt or --decrypt");
                System.exit(1);
            }

            if (cmdline.hasOption("encrypt")) {
                direction = CryptoOp.ENCRYPT;
            } else {
                direction = CryptoOp.DECRYPT;
            }

            if (!cmdline.hasOption("key-id") &&
                !cmdline.hasOption("key-name")) {
                help("You must specify an encryption key with --key-id " +
                     "or --key-name");
                System.exit(1);
            }

            if (cmdline.hasOption("key-id") && cmdline.hasOption("key-name")) {
                System.err.println("Both --key-id and --key-name were " +
                                   "provided");
                System.err.println("Using the key name and ignoring id");
            }

            if (cmdline.hasOption("key-id")) {
                keyId = cmdline.getOptionValue("key-id");
            } else if (cmdline.hasOption("key-name")) {
                keyName = cmdline.getOptionValue("key-name");
            }

            if (direction == CryptoOp.ENCRYPT &&
                !cmdline.hasOption("iv-in") &&
                !cmdline.hasOption("iv-out")) {
                help("When encrypting, either --iv-in or --iv-out is " +
                     "required. If --iv-in is not provided, the " +
                     "initialization value will be generated by the SDKMS " +
                     "server and written to the " +
                     "filename specified by --iv-out");
                System.exit(1);
            }
            if (direction == CryptoOp.DECRYPT &&
                (!cmdline.hasOption("iv-in") ||
                 cmdline.hasOption("iv-out"))) {
                help("When decrypting, --iv-in is required and --iv-out " +
                     "cannot be specified");
                System.exit(1);
            }

            if (cmdline.hasOption("iv-in")) {
                ivIn = cmdline.getOptionValue("iv-in");
            }
            if (cmdline.hasOption("iv-out")) {
                ivOut = cmdline.getOptionValue("iv-out");
            }

            if (direction == CryptoOp.ENCRYPT) {
                if (!cmdline.hasOption("tag-len")) {
                    help("When encrypting, --tag-len is required");
                    System.exit(1);
                }
                tagLen = new Integer(
                    ((Number) cmdline.getParsedOptionValue("tag-len")).
                    intValue());
                if (!cmdline.hasOption("tag-out")) {
                    help("When encrypting, --tag-out is required");
                    System.exit(1);
                }
                tagOut = cmdline.getOptionValue("tag-out");
                if (cmdline.hasOption("tag-in")) {
                    help("When encrypting, --tag-in cannot be specified");
                    System.exit(1);
                }
            } else {
                if (cmdline.hasOption("tag-len")) {
                    help("When decrypting, --tag-len cannot be specified");
                    System.exit(1);
                }
                if (cmdline.hasOption("tag-out")) {
                    help("When decrypting, --tag-out cannot be specified");
                    System.exit(1);
                }
                if (!cmdline.hasOption("tag-in")) {
                    help("When decrypting, --tag-in is required");
                    System.exit(1);
                }
                tagIn = cmdline.getOptionValue("tag-in");
            }

            if (cmdline.hasOption("ad-in")) {
                adIn = cmdline.getOptionValue("ad-in");
            }

            if (cmdline.hasOption("tag-out")) {
                tagOut = cmdline.getOptionValue("tag-out");
            }
        } catch (ParseException e) {
            usage(options);
            System.err.println("Command-line parsing failed: " +
                               e.getMessage());
            System.exit(1);
        }
    }

    public static void main(String[] args) throws ApiException {
        int status = 0;

        AesGcmCrypto crypto = new AesGcmCrypto();
        crypto.parseCmdline(args);
        crypto.connect();
        try {
            crypto.performCrypto();
        } catch (ApiException | IOException e) {
            System.err.println("Crypto failed: " + e.getMessage());
            status = 1;
        } finally {
            crypto.logout();
        }
        System.exit(status);
    }
}
